import typing as t

from py_ecc.bls import G2ProofOfPossession as bls
from lido.eth2deposit.utils.ssz import (
    DepositMessage,
    compute_deposit_domain,
    compute_signing_root,
)
from lido.eth2deposit.settings import get_chain_setting
from lido.constants.chains import get_eth2_chain_name
from lido.contracts.w3_contracts import get_lido_contract

import concurrent

# Prepare network
lido = get_lido_contract()
withdrawal_credentials = bytes(lido.functions.getWithdrawalCredentials().call())
fork_version = get_chain_setting(get_eth2_chain_name()).GENESIS_FORK_VERSION
domain = compute_deposit_domain(fork_version=fork_version)


def validate_key(key: t.Dict) -> bool:
    """Run signature validation on a key"""

    # Is this key already validated? eg in cache
    if "valid_signature" in key.keys():
        return

    pubkey = bytes.fromhex(key["key"]) if type(key["key"]) is str else key["key"]
    signature = (
        bytes.fromhex(key["depositSignature"])
        if type(key["depositSignature"]) is str
        else key["depositSignature"]
    )

    # Minimum staking requirement of 32 ETH per validator
    REQUIRED_DEPOSIT_ETH = 32
    ETH2GWEI = 10 ** 9
    amount = REQUIRED_DEPOSIT_ETH * ETH2GWEI

    deposit_message = DepositMessage(
        pubkey=pubkey,
        withdrawal_credentials=withdrawal_credentials,
        amount=amount,
    )

    signing_root = compute_signing_root(deposit_message, domain)

    return bls.Verify(pubkey, signing_root, signature)


def validate_keys_mono(operators):
    """This is an additional, single-process key validation function"""

    for op_i, op in enumerate(operators):
        for key_i, key in enumerate(op["keys"]):
            # Is this key already validated? eg in cache
            if "valid_signature" not in key.keys():
                operators[op_i]["keys"][key_i]["valid_signature"] = validate_key(key)

    return operators


def validate_keys_multi(operators):
    """
    Main multi-process validation function.
    Modifies the input! Adds "valid_signature" field to every key item.
    It will spawn an appropriate process pool for the amount of threads on processor.
    """

    with concurrent.futures.ProcessPoolExecutor() as executor:
        for op_i, op in enumerate(operators):
            validate_key_results = executor.map(validate_key, op["keys"])
            for key_index, validate_key_result in enumerate(validate_key_results):
                # Is this key already validated? eg in cache
                if validate_key_result is not None:
                    operators[op_i]["keys"][key_index]["valid_signature"] = validate_key_result

    return operators



from web3.auto import w3
from lido.constants.contract_addresses import (
    get_lido_address,
    get_registry_address,
)
from lido.contracts.abi_loader import load_lido_abi, load_operators_abi


def get_lido_contract():
    """Load Lido web3 contract object"""
    return w3.eth.contract(address=get_lido_address(), abi=load_lido_abi())


def get_nos_contract():
    """Load Node Operator web3 contract object"""
    return w3.eth.contract(address=get_registry_address(), abi=load_operators_abi())



from web3.auto import w3

# Inject middleware for geth if we are on Goerli
if w3.eth.chainId == 5:
    from web3.middleware import geth_poa_middleware

    w3.middleware_onion.inject(geth_poa_middleware, layer=0)


def get_data_actuality():
    return {
        "last_block": w3.eth.getBlock("latest")["number"],
        "last_blocktime": w3.eth.getBlock("latest")["timestamp"],
    }



from typing import List

from web3.auto import w3

from lido.multicall import Call
from lido.multicall.constants import MULTICALL_ADDRESSES


class Multicall:
    def __init__(self, calls: List[Call]):
        self.calls = calls

    def __call__(self):
        aggregate = Call(
            MULTICALL_ADDRESSES[w3.eth.chainId],
            "aggregate((address,bytes)[])(uint256,bytes[])",
        )
        args = [[[call.target, call.data] for call in self.calls]]
        block, outputs = aggregate(args)
        result = {}
        for call, output in zip(self.calls, outputs):
            result.update(call.decode_output(output))
        return result



from enum import IntEnum


class Network(IntEnum):
    Mainnet = 1
    Kovan = 42
    Rinkeby = 4
    Görli = 5
    xDai = 100


MULTICALL_ADDRESSES = {
    Network.Mainnet: '0xeefBa1e63905eF1D7ACbA5a8513c70307C1cE441',
    Network.Kovan: '0x2cc8688C5f75E365aaEEb4ea8D6a480405A48D2A',
    Network.Rinkeby: '0x42Ad527de7d4e9d9d011aC45B31D8551f8Fe9821',
    Network.Görli: '0x77dCa2C955b15e9dE4dbBCf1246B4B85b651e50e',
    Network.xDai: '0xb5b692a88BDFc81ca69dcB1d924f59f0413A602a',
}



import os


ZERO_BYTES32 = b"\x00" * 32

# Eth2-spec constants taken from https://github.com/ethereum/eth2.0-specs/blob/dev/specs/phase0/beacon-chain.md
DOMAIN_DEPOSIT = bytes.fromhex("03000000")
BLS_WITHDRAWAL_PREFIX = bytes.fromhex("00")

ETH2GWEI = 10 ** 9
MIN_DEPOSIT_AMOUNT = 2 ** 0 * ETH2GWEI
MAX_DEPOSIT_AMOUNT = 2 ** 5 * ETH2GWEI


# File/folder constants
WORD_LISTS_PATH = os.path.join("eth2deposit", "key_handling", "key_derivation", "word_lists")
DEFAULT_VALIDATOR_KEYS_FOLDER_NAME = "validator_keys"


# Sundry constants
UNICODE_CONTROL_CHARS = list(range(0x00, 0x20)) + list(range(0x7F, 0xA0))



import typing as t

from lido import (
    get_operators_data,
    get_operators_keys,
    validate_keys_multi,
    find_duplicates,
)


def fetch_and_validate(
    registry_address: t.Optional[str] = None, max_multicall: int = 300
) -> t.List[t.Dict]:
    """fetches all operator keys, and run keys validations and duplication checks

    check fields:
     - operators[op_i]["keys"][key_i]["duplicate"]
     - operators[op_i]["keys"][key_i]["duplicates"]
     - operators[op_i]["keys"][key_i]["valid_signature"]
    """
    operator_keys = get_operators_keys(
        get_operators_data(registry_address=registry_address), max_multicall=max_multicall
    )
    result = find_duplicates(validate_keys_multi(operator_keys))
    return result



import typing as t

from lido.multicall import Call, Multicall
from lido.constants.contract_addresses import get_registry_address
from lido.contracts.w3_contracts import get_nos_contract


def get_operators_data(registry_address: t.Optional[str] = None):
    """Fetch information for each node operator"""

    address = registry_address or get_registry_address()

    operators_n = Call(address, "getNodeOperatorsCount()(uint256)")()

    calls = Multicall(
        [
            Call(
                address,
                [
                    "getNodeOperator(uint256,bool)(bool,string,address,uint64,uint64,uint64,uint64)",
                    i,
                    True,
                ],
                [[i, None]],
            )
            for i in range(operators_n)
        ]
    )()

    calls_list = list(calls.values())

    # Adding index as first data of operator
    calls_with_indeces = [[i] + list(item) for i, item in enumerate(calls_list)]

    # Getting function data from contract ABI
    function_abi = next(x for x in get_nos_contract().abi if x["name"] == "getNodeOperator")

    # Adding "id" and the rest of output name keys
    op_keys = ["id"] + [x["name"] for x in function_abi["outputs"]]
    operators = [dict(zip(op_keys, op)) for op in calls_with_indeces]

    return operators



from web3.auto import w3

# We are deployed on Goerli and Mainnet
chains = {1: "mainnet", 5: "goerli"}


def get_chain_name():
    """Return a readable network name for current network"""
    return chains[w3.eth.chainId]


def get_eth2_chain_name():
    """Return a readable network name for current network for eth2-deposit-cli library"""

    # Return Pyrmont for Goerli deployment
    if w3.eth.chainId == 5:
        return "pyrmont"

    return chains[w3.eth.chainId]



from lido.multicall import Call, Multicall
from lido.constants.contract_addresses import get_lido_address
from lido.contracts.w3_contracts import get_lido_contract
from lido.utils.data_actuality import get_data_actuality

# Default functions to execute, this can be easily overridden
funcs_to_fetch = [
    "isStopped",
    "getTotalPooledEther",
    "getWithdrawalCredentials",
    "getFee",
    "getFeeDistribution",
    "getBeaconStat",
    "getBufferedEther",
]


def get_stats(funcs_to_fetch=funcs_to_fetch):
    """Fetch various constants from Lido for analytics and statistics"""

    address = get_lido_address()

    # Getting function data from contract ABI
    funcs_from_contract = [
        x
        for x in get_lido_contract().abi
        if x["type"] == "function" and x["name"] in funcs_to_fetch
    ]

    # Adding "multicall_outputs" with prepared input data for multicall
    for func_i, func in enumerate(funcs_from_contract):
        x = []
        for output in func["outputs"]:
            x.append(output["type"])
        funcs_from_contract[func_i]["multicall_outputs"] = ",".join(x)

    calls = Multicall(
        [
            Call(
                address,
                [
                    "%s()(%s)" % (item["name"], item["multicall_outputs"]),
                ],
                [[item["name"], None]],
            )
            for item in funcs_from_contract
        ]
    )()

    # Return values instead of single-element tuples
    for call in calls:
        item = calls[call]
        if type(item) == tuple and len(item) == 1:
            calls[call] = item[0]

    actuality_data = get_data_actuality()

    return {**actuality_data, **calls}



__version__ = "0.1.1"

from lido.get_operators_data import get_operators_data  # noqa: F401
from lido.get_operators_keys import get_operators_keys  # noqa: F401
from lido.validate_keys import validate_key  # noqa: F401
from lido.validate_keys import validate_keys_mono  # noqa: F401
from lido.validate_keys import validate_keys_multi  # noqa: F401
from lido.find_duplicates import find_duplicates  # noqa: F401
from lido.fetch_and_validate import fetch_and_validate  # noqa: F401
from lido.get_stats import get_stats  # noqa: F401
from lido.beacon import get_beacon  # noqa: F401
from lido.utils.data_actuality import get_data_actuality  # noqa: F401



from eth_utils import to_checksum_address
from web3.auto import w3
from lido.multicall import Signature


class Call:
    def __init__(self, target, function, returns=None):
        self.target = to_checksum_address(target)
        if isinstance(function, list):
            self.function, *self.args = function
        else:
            self.function = function
            self.args = None
        self.signature = Signature(self.function)
        self.returns = returns

    @property
    def data(self):
        return self.signature.encode_data(self.args)

    def decode_output(self, output):
        decoded = self.signature.decode_data(output)

        if self.returns:
            return {
                name: handler(value) if handler else value
                for (name, handler), value in zip(self.returns, [decoded])
            }
        else:
            return decoded if len(decoded) > 1 else decoded[0]

    def __call__(self, args=None):
        args = args or self.args
        calldata = self.signature.encode_data(args)
        output = w3.eth.call({"to": self.target, "data": calldata})
        return self.decode_output(output)



import os
import json
from lido.constants.chains import get_chain_name

network = get_chain_name()
script_dir = os.path.dirname(__file__)


def load_lido_abi():
    """Load an appropriate ABI file for Lido"""

    env = os.getenv("lido_abi")

    if env is not None:
        return json.load(open(env))

    if network == "mainnet":
        return json.load(open(os.path.join(script_dir, "abi/mainnet/Lido.json")))
    elif network == "goerli":
        return json.load(open(os.path.join(script_dir, "abi/goerli/Lido.json")))
    else:
        raise Exception("Unable to load Lido ABI for network")


def load_operators_abi():
    """Load an appropriate ABI file for Node Operators"""

    env = os.getenv("REGISTRY_ABI")

    if env is not None:
        return json.load(open(env))

    if network == "mainnet":
        return json.load(
            open(os.path.join(script_dir, "abi/mainnet/NodeOperatorsRegistry.json"))
        )
    elif network == "goerli":
        return json.load(
            open(os.path.join(script_dir, "abi/goerli/NodeOperatorsRegistry.json"))
        )
    else:
        raise Exception("Unable to load NodeOperatorsRegistry ABI for network")
